#!/usr/bin/env python

# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-21  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Provide a couple of :py:class:`datetime.tzinfo` compatible objects
representing local time and UTC.

Introduction
------------

This module provides two :py:class:`datetime.tzinfo` compatible objects
representing UTC and local time zones. These are used to convert
timestamps to and from UTC and local time. The weather station software
stores data with UTC timestamps, to avoid problems with daylight savings
time, but the template and plot programs output data with local times.

Detailed API
------------

"""

from __future__ import print_function

import datetime
import logging
import sys

import tzlocal

from pywws.constants import DAY, HOUR

logger = logging.getLogger(__name__)


class _TimeZone(object):
    class _UTC(datetime.tzinfo):
        _offset = datetime.timedelta(0)

        def utcoffset(self, dt):
            return self._offset

        def dst(self, dt):
            return self._offset

        def tzname(self, dt):
            return 'UTC'


    def __init__(self, tz_name=None):
        if tz_name:
            # use a named time zone instead of system default, for testing
            import pytz
            self.local = pytz.timezone(tz_name)
        else:
            self.local = tzlocal.get_localzone()
        logger.info('Using timezone "{!s}"'.format(self.local))
        self._using_pytz = hasattr(self.local, 'localize')
        if sys.version_info >= (3, 2):
            self.utc = datetime.timezone.utc
        else:
            self.utc = self._UTC()

    def local_to_utc(self, dt):
        """Convert a local time (with or without tzinfo) to UTC without
        tzinfo, as used for pywws timestamps."""
        if dt.tzinfo is None:
            if self._using_pytz:
                dt = self.local.localize(dt)
            else:
                dt = dt.replace(tzinfo=self.local)
        return dt.astimezone(self.utc).replace(tzinfo=None)

    def utc_to_local(self, dt):
        """Convert a pywws time to local time, with tzinfo."""
        return dt.replace(tzinfo=self.utc).astimezone(self.local)

    def utc_to_nodst(self, dt):
        """Convert a pywws time to local time, without DST.

        This could be an invalid time, so no tzinfo is included."""
        dt = self.utc_to_local(dt)
        return dt.replace(tzinfo=None) - dt.dst()

    def hour_start(self, dt):
        """Return UTC time before dt whose local time is on the hour."""
        local = self.utc_to_local(dt)
        diff = local - local.replace(minute=0, second=0)
        return dt - diff

    def day_start(self, dt, day_end_hour, use_dst=True):
        """Return UTC time before dt whose local time is day_end_hour.

        If use_dst is False then DST is ignored, so a day ending at 9am
        (local time) in winter would end at 10am in summer."""
        # get UTC and local equivalent
        dt = dt.replace(tzinfo=self.utc)
        local = dt.astimezone(self.local)
        if not use_dst:
            local = local.replace(tzinfo=None) - local.dst()
        # truncate to the hour
        diff = local - local.replace(minute=0, second=0)
        dt -= diff
        local -= diff
        # decrement UTC until equivalent local time has correct hour
        for h in range(4):
            hours = local.hour - day_end_hour
            if hours == 0:
                break
            elif hours > 1:
                hours -= 1
            elif hours < 0:
                hours += 23
            dt -= HOUR * hours
            local = dt.astimezone(self.local)
            if not use_dst:
                local = local.replace(tzinfo=None) - local.dst()
        return dt.replace(tzinfo=None)

    def _dst_transitions(self):
        """Find local times near DST transitions in the last year.

        This is only intended to be used for testing."""
        day = datetime.datetime.now(tz=self.local).replace(
            minute=15, second=0, microsecond=0)
        dst = day.dst()
        for d in range(365):
            day -= DAY
            if self._using_pytz:
                day = self.local.normalize(day)
            if day.dst() == dst:
                continue
            hour = day
            for h in range(25):
                hour += HOUR
                if self._using_pytz:
                    hour = self.local.normalize(hour)
                if hour.dst() == dst:
                    yield hour
                    break
            dst = day.dst()


time_zone = _TimeZone()


def main():
    global time_zone
    if len(sys.argv) > 1:
        time_zone = _TimeZone(sys.argv[1])
    print('Local time zone:', time_zone.local)
    now = datetime.datetime.utcnow().replace(microsecond=0)
    print('UTC time:', now)
    lcl = time_zone.utc_to_local(now)
    print('Local time:', lcl, lcl.strftime('(%Z)'))
    print()
    print('most recent 9am', time_zone.day_start(now, 9), 'UTC')
    print('most recent 9pm', time_zone.day_start(now, 21), 'UTC')
    for dt in time_zone._dst_transitions():
        print()
        print('Convert local to UTC around DST transition')
        lcl = dt.replace(tzinfo=None)
        lcl -= HOUR * 2
        for h in range(4):
            utc = time_zone.local_to_utc(lcl)
            print(lcl, '->', utc)
            lcl += HOUR
        print()
        print('Convert UTC to local around DST transition')
        utc -= HOUR * 3
        for h in range(4):
            lcl = time_zone.utc_to_local(utc)
            print(utc, '->', lcl, lcl.strftime('(%Z)'))
            utc += HOUR
    return 0


if __name__ == "__main__":
    sys.exit(main())



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-18  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Copy files to another directory.

This module can be used to copy template and graph results to another
directory on your computer. This could be useful if you are running a
web server on the same machine as pywws (or on a machine that's
accessible as a network share).

* Example ``weather.ini`` configuration::

    [copy]
    directory = /home/www/public_html/weather/data/

    [hourly]
    plot = ['24hrs.png.xml', 'rose_12hrs.png.xml']
    text = ['24hrs.txt']
    services = [('copy', '24hrs.txt', '24hrs.png', 'rose_12hrs.png')]

Run :py:mod:`pywws.service.copy` once to set the default configuration,
which you can then change. ``directory`` is the full path of a directory
in which all the copied files will be put.

You can copy any files you like, as often as you like, but typical usage
is to update a website once an hour. Each file to be uploaded needs to
be listed in a service entry like ``('copy', 'filename')``. If the file
is not in your ``work`` directory's ``output`` directory then
``filename`` should be the full path.

If you need to copy some files to a different directory you can copy the
:py:mod:`pywws.service.copy` module to your ``modules`` directory,
making sure that you rename it, for example to ``copy2.py``. This
creates a ``copy2`` service that you can use in the same way, but with a
different ``directory``.

"""

from __future__ import absolute_import

from contextlib import contextmanager
from datetime import timedelta
import logging
import os
import shutil
import sys

import pywws.service

__docformat__ = "restructuredtext en"
service_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(__name__)


class ToService(pywws.service.FileService):
    config = {'directory'  : ('', True,  None)}
    logger = logger
    service_name = service_name

    @contextmanager
    def session(self):
        yield None

    def upload_file(self, session, path):
        try:
            if not os.path.isdir(self.params['directory']):
                os.makedirs(self.params['directory'])
            shutil.copy(path, self.params['directory'])
        except Exception as ex:
            return False, repr(ex)
        return True, 'OK'


if __name__ == "__main__":
    sys.exit(pywws.service.main(ToService))



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-21  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Predict future weather using recent data
::

%s

"""

from __future__ import absolute_import, print_function

__docformat__ = "restructuredtext en"
__usage__ = """
 usage: python -m pywws.forecast [options] data_dir
 options are:
  -h | --help  display this help
 data_dir is the root directory of the weather data
"""
__doc__ %= __usage__
__usage__ = __doc__.split('\n')[0] + __usage__

from ast import literal_eval
from datetime import datetime, timedelta
import getopt
import sys

import pywws.localisation
import pywws.storage

def _(msg):
    return msg

_forecast_text = {
    'A' : _("Settled fine"),
    'B' : _("Fine weather"),
    'C' : _("Becoming fine"),
    'D' : _("Fine, becoming less settled"),
    'E' : _("Fine, possible showers"),
    'F' : _("Fairly fine, improving"),
    'G' : _("Fairly fine, possible showers early"),
    'H' : _("Fairly fine, showery later"),
    'I' : _("Showery early, improving"),
    'J' : _("Changeable, mending"),
    'K' : _("Fairly fine, showers likely"),
    'L' : _("Rather unsettled clearing later"),
    'M' : _("Unsettled, probably improving"),
    'N' : _("Showery, bright intervals"),
    'O' : _("Showery, becoming less settled"),
    'P' : _("Changeable, some rain"),
    'Q' : _("Unsettled, short fine intervals"),
    'R' : _("Unsettled, rain later"),
    'S' : _("Unsettled, some rain"),
    'T' : _("Mostly very unsettled"),
    'U' : _("Occasional rain, worsening"),
    'V' : _("Rain at times, very unsettled"),
    'W' : _("Rain at frequent intervals"),
    'X' : _("Rain, very unsettled"),
    'Y' : _("Stormy, may improve"),
    'Z' : _("Stormy, much rain")
    }

del _


def zambretti_code(params, hourly_data):
    """Simple implementation of Zambretti forecaster algorithm.
    Inspired by beteljuice.com Java algorithm, as converted to Python by
    honeysucklecottage.me.uk, and further information
    from http://www.meteormetrics.com/zambretti.htm"""
    north = literal_eval(params.get('Zambretti', 'north', 'True'))
    baro_upper = float(params.get('Zambretti', 'baro upper', '1050.0'))
    baro_lower = float(params.get('Zambretti', 'baro lower', '950.0'))
    if not hourly_data['rel_pressure']:
        return ''
    if hourly_data['wind_ave'] is None or hourly_data['wind_ave'] < 0.3:
        wind = None
    else:
        wind = hourly_data['wind_dir']
    if hourly_data['pressure_trend'] is None:
        trend = 0.0
    else:
        trend = hourly_data['pressure_trend'] / 3.0
    # normalise pressure
    pressure = 950.0 + (
        (1050.0 - 950.0) * (hourly_data['rel_pressure'] - baro_lower) /
        (baro_upper - baro_lower))
    # adjust pressure for wind direction
    if wind is not None:
        if not isinstance(wind, int):
            wind = int(wind + 0.5) % 16
        if not north:
            # southern hemisphere, so add 180 degrees
            wind = (wind + 8) % 16
        pressure += (  5.2,  4.2,  3.2,  1.05, -1.1, -3.15, -5.2, -8.35,
                     -11.5, -9.4, -7.3, -5.25, -3.2, -1.15,  0.9,  3.05)[wind]
    # compute base forecast from pressure and trend (hPa / hour)
    summer = north == (hourly_data['idx'].month >= 4 and
                       hourly_data['idx'].month <= 9)
    if trend >= 0.1:
        # rising pressure
        if summer:
            pressure += 3.2
        F = 0.1740 * (1031.40 - pressure)
        LUT = ('A', 'B', 'B', 'C', 'F', 'G', 'I', 'J', 'L', 'M', 'M', 'Q', 'T',
               'Y')
    elif trend <= -0.1:
        # falling pressure
        if summer:
            pressure -= 3.2
        F = 0.1553 * (1029.95 - pressure)
        LUT = ('B', 'D', 'H', 'O', 'R', 'U', 'V', 'X', 'X', 'Z')
    else:
        # steady
        F = 0.2314 * (1030.81 - pressure)
        LUT = ('A', 'B', 'B', 'B', 'E', 'K', 'N', 'N', 'P', 'P', 'S', 'W', 'W',
               'X', 'X', 'X', 'Z')
    # clip to range of lookup table
    F = min(max(int(F + 0.5), 0), len(LUT) - 1)
    # convert to letter code
    return LUT[F]


def zambretti(params, hourly_data):
    code = zambretti_code(params, hourly_data)
    return pywws.localisation.translation.ugettext(_forecast_text[code])


def main(argv=None):
    from pywws.timezone import time_zone
    if argv is None:
        argv = sys.argv
    try:
        opts, args = getopt.getopt(argv[1:], "h", ['help'])
    except getopt.error as msg:
        print('Error: %s\n' % msg, file=sys.stderr)
        print(__usage__.strip(), file=sys.stderr)
        return 1
    # process options
    for o, a in opts:
        if o in ('-h', '--help'):
            print(__usage__.strip())
            return 0
    # check arguments
    if len(args) != 1:
        print("Error: 1 argument required", file=sys.stderr)
        print(__usage__.strip(), file=sys.stderr)
        return 2
    data_dir = args[0]
    with pywws.storage.pywws_context(data_dir) as context:
        params = context.params
        pywws.localisation.set_application_language(params)
        hourly_data = context.hourly_data
        idx = hourly_data.before(datetime.max)
        print('Zambretti (current):', zambretti(params, hourly_data[idx]))
        idx = time_zone.utc_to_local(idx)
        if idx.hour < 8 or (idx.hour == 8 and idx.minute < 30):
            idx -= timedelta(hours=24)
        idx = idx.replace(hour=9, minute=0, second=0)
        idx = time_zone.local_to_utc(idx)
        idx = hourly_data.nearest(idx)
        lcl = time_zone.utc_to_local(idx)
        print('Zambretti (at %s):' % lcl.strftime('%H:%M %Z'), zambretti(
            params, hourly_data[idx]))
    return 0

if __name__ == "__main__":
    sys.exit(main())



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2018-20  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Base classes for "service" uploaders.

.. inheritance-diagram:: CatchupDataService FileService LiveDataService
    :top-classes: pywws.service.ServiceBase

"""

from __future__ import absolute_import, print_function, unicode_literals

from ast import literal_eval
from collections import deque
from datetime import datetime, timedelta
import os
import sys
import threading

if sys.version_info[0] >= 3:
    from io import StringIO
else:
    from StringIO import StringIO

import pywws
from pywws.constants import SECOND
import pywws.logger
import pywws.storage
import pywws.template


class Queue(deque):
    def __init__(self, start, *arg, **kw):
        super(Queue, self).__init__(*arg, **kw)
        self._start = start

    def append(self, x):
        super(Queue, self).append(x)
        if x is None:
            return
        self.append = super(Queue, self).append
        self._start()

    def full(self):
        """Are there already too many uploads on the queue."""
        return len(self) >= 50


class ServiceBase(threading.Thread):
    """Base class for all service uploaders.

    Uploaders use a separate thread to allow the main program thread to
    continue even if a service is slow to respond. Items to upload are
    passed to the thread via a thread safe queue. The thread is started
    when the first item is put on the queue. To shut down the thread put
    :py:obj:`None` on the queue, e.g. by calling :py:meth:`stop`.

    There are two types of uploader derived from this class.
    :py:class:`DataServiceBase` is used by uploaders that send defined
    sets of data, typically as an HTML "post" or "get" operation.
    :py:class:`FileService` is used to upload files, including free form
    text such as a Twitter message.

    All service classes must provide a :py:attr:`logger` object so that
    logging messages carry the right module name, and define a
    :py:attr:`service_name` string. They must also define a
    :py:meth:`session` method.
    """

    config = {}
    """Defines the user configuration of the uploader. Each item must be
    of the form ``name: (default (str), required (bool), fixed_key (str
    or None))``. ``name`` is the ``weather.ini`` value name, ``default``
    is a default value, ``required`` defines whether a value must be
    supplied at run time, and ``fixed_key`` defines if and to where in
    :py:attr:`~DataServiceBase.fixed_data` the value should be copied.
    """

    interval = timedelta(seconds=40)
    """Sets the minimum period between the timestamps of uploaded data.
    For some services this can be less than the weather station's "live"
    data period (48 seconds) whereas others may require 5 or 15 minutes
    between readings.
    """

    logger = None
    """A :py:class:`logging.Logger` object created with the module name.
    This is typically done as follows::

        logger = logging.getLogger(__name__)
    """

    service_name = ''
    """A short name used to refer to the service in weather.ini. It
    should be all lower case. The best name to use is the last part of
    the module's file name, as follows::

        service_name = os.path.splitext(os.path.basename(__file__))[0]
    """

    def __init__(self, context, check_params=True):
        super(ServiceBase, self).__init__()
        self.context = context
        self.queue = Queue(self.start)
        # get user configuration
        self.params = {}
        check = []
        for key, (default, required, fixed_key) in self.config.items():
            self.params[key] = context.params.get(
                self.service_name, key, default)
            if required:
                check.append(key)
            if fixed_key and self.params[key]:
                # copy fixed_data to avoid changing class definition
                self.fixed_data = dict(self.fixed_data)
                self.fixed_data[fixed_key] = self.params[key]
        # check values
        if check_params:
            self.check_params(*check)

    def check_params(self, *keys):
        """Ensure user has set required values in weather.ini.

        Normally the :py:data:`~ServiceBase.config` names with
        ``required`` set are checked, but if your uploader has a
        ``register`` method you may need to check for other data.

        :param str keys: the :py:data:`~ServiceBase.config` names to
            verify.
        """

        for key in keys:
            if not self.params[key]:
                raise RuntimeError('"{}" not set in weather.ini'.format(key))

    def session(self):
        """Context manager factory function for a batch of one or more
        uploads.

        This makes it easy to ensure any resources such as an internet
        connection are properly closed after a batch of uploads. Use the
        :py:func:`contextlib.contextmanager` decorator when you
        implement this method.

        For a typical example, see the source code of the
        :py:mod:`pywws.service.openweathermap` module. If your upload
        can't benefit from a session object yield :py:obj:`None`, as in
        :py:mod:`pywws.service.copy`.
        """
        raise NotImplementedError()

    def run(self):
        """ """
        self.logger.debug('thread started ' + self.name)
        self.old_message = ''
        if self.context.live_logging:
            polling_interval = self.interval.total_seconds() / 20
            polling_interval = min(max(polling_interval, 4.0), 40.0)
        else:
            polling_interval = 4.0
        while not self.context.shutdown.is_set():
            OK = True
            if self.queue:
                try:
                    OK = self.upload_batch()
                except Exception as ex:
                    self.logger.exception(ex)
                    OK = False
            if OK:
                pause = polling_interval
            elif self.context.live_logging:
                # upload failed, wait before trying again
                pause = 40.0
            else:
                # upload failed or nothing more to do
                break
            self.context.shutdown.wait(pause)

    def stop(self):
        if self.is_alive():
            self.logger.debug('stopping thread ' + self.name)
            self.queue.append(None)

    def log(self, message):
        if message == self.old_message:
            self.logger.debug(message)
        else:
            self.logger.error(message)
            self.old_message = message


class DataServiceBase(ServiceBase):
    """Base class for "data" services.

    A "data" service uploader sends defined sets of data, typically as
    an HTML "post" or "get" operation. Service classes should be based
    on :py:class:`CatchupDataService` or :py:class:`LiveDataService`,
    depending on whether the service allows uploading of past data, for
    example to fill in gaps if the server (or pywws client) goes down
    for a few hours or days.

    Data service classes must provide a :py:attr:`template` string to
    define how to convert pywws data before uploading. Required methods
    are :py:meth:`~ServiceBase.session` and :py:meth:`upload_data`. If
    the service has a separate authorisation or registration process
    this can be done in a :py:meth:`~pywws.service.mastodon.register`
    method. See :py:mod:`pywws.service.mastodon` for an example.
    """

    template = ''
    """Defines the conversion of pywws data to key, value pairs required
    by the service. The template string is passed to
    :py:mod:`pywws.template`, then the result is passed to
    :py:func:`~ast.literal_eval` to create a :py:obj:`dict`. This rather
    complex process allows great flexibility, but you do have to be
    careful with use of quotation marks. """

    fixed_data = {}
    """Defines a set of ``key: value`` pairs that are the same for every
    data upload. This might include the station's location or the
    software name & version. Values set by the user should be included
    in the weather.ini config defined in :py:data:`~ServiceBase.config`.
    """

    def __init__(self, context, check_params=True):
        super(DataServiceBase, self).__init__(context, check_params)
        # check config
        template = context.params.get(self.service_name, 'template')
        if template == 'default':
            context.params.unset(self.service_name, 'template')
        elif template:
            self.logger.critical(
                'obsolete item "template" found in weather.ini '
                'section [{}]'.format(self.service_name))
        # create templater
        if self.template:
            self.templater = pywws.template.Template(context, use_locale=False)
            self.template_file = None
        # get time stamp of last uploaded data
        self.last_update = context.status.get_datetime(
            'last update', self.service_name)
        if not self.last_update:
            if self.catchup:
                self.last_update = datetime.utcnow() - timedelta(
                    days=self.catchup)
            else:
                self.last_update = datetime.min

    def upload_data(self, session, prepared_data={}):
        """Upload one data set to the service.

        Every data service class must implement this method.

        :param object session: the object created by
            :py:meth:`~ServiceBase.session`. This is typically used to
            communicate with the server and is automatically closed when
            a batch of uploads has finished.
        :param dict prepared_data: a set of key: value pairs to upload.
            The keys and values must all be text strings.
        """
        raise NotImplementedError()

    def queue_data(self, timestamp, data):
        if not self.valid_data(data):
            return False
        prepared_data = self.prepare_data(data)
        prepared_data.update(self.fixed_data)
        self.logger.debug('data: %s', str(prepared_data))
        self.queue.append((timestamp, prepared_data))
        return True

    def prepare_data(self, data):
        if not self.template_file:
            self.template_file = StringIO(self.template)
        data_str = self.templater.make_text(self.template_file, data)
        self.template_file.seek(0)
        return literal_eval('{' + data_str + '}')

    def valid_data(self, data):
        return True


class CatchupDataService(DataServiceBase):
    catchup = 7
    """Sets the number of days of past data that can be uploaded when a
    service is first used.
    """

    def queue_data(self, timestamp, data):
        if timestamp and timestamp < self.last_update + self.interval:
            return False
        OK = super(CatchupDataService, self).queue_data(timestamp, data)
        if OK and timestamp:
            self.last_update = timestamp
        return OK

    def do_catchup(self, do_all=False):
        start = self.last_update + self.interval
        if do_all:
            for data in self.context.calib_data[start:]:
                while self.queue.full():
                    self.context.shutdown.wait(4.0)
                    if self.context.shutdown.is_set():
                        return True
                self.queue_data(data['idx'], data)
            return True
        for data in self.context.calib_data[start:]:
            if self.queue.full():
                return True
            if self.queue_data(data['idx'], data):
                return False
        return True

    def upload(self, live_data=None, test_mode=False, options=()):
        if self.queue.full():
            return
        if test_mode:
            start = self.context.calib_data.before(datetime.max)
        else:
            start = self.last_update + self.interval
        for data in self.context.calib_data[start:]:
            timestamp = data['idx']
            if test_mode:
                timestamp = None
            if self.queue_data(timestamp, data):
                return
        if live_data:
            self.queue_data(live_data['idx'], live_data)

    def upload_batch(self):
        OK = True
        count = 0
        with self.session() as session:
            while self.queue and not self.context.shutdown.is_set():
                # send upload without taking it off queue
                upload = self.queue[0]
                if upload is None:
                    OK = False
                    break
                timestamp, prepared_data = upload
                OK, message = self.upload_data(
                    session, prepared_data=prepared_data)
                self.log(message)
                if not OK:
                    break
                count += 1
                if timestamp:
                    self.context.status.set(
                        'last update', self.service_name, str(timestamp))
                # finally remove upload from queue
                self.queue.popleft()
        if count > 1:
            self.logger.warning('{:d} records sent'.format(count))
        elif count:
            self.logger.info('1 record sent')
        return OK


class LiveDataService(DataServiceBase):
    catchup = None

    def do_catchup(self, do_all=False):
        return True

    def upload(self, live_data=None, test_mode=False, options=()):
        if self.queue.full():
            return
        if live_data:
            data = live_data
        else:
            idx = self.context.calib_data.before(datetime.max)
            if not idx:
                return
            data = self.context.calib_data[idx]
        timestamp = data['idx']
        if test_mode:
            timestamp = None
        self.queue_data(timestamp, data)

    def upload_batch(self):
        # get most recent upload on queue
        upload = self.queue.popleft()
        while self.queue and self.queue[0] is not None:
            upload = self.queue.popleft()
        if upload is None:
            return False
        timestamp, prepared_data = upload
        # check time since last upload
        if timestamp and timestamp < self.last_update + self.interval:
            return True
        with self.session() as session:
            OK, message = self.upload_data(session, prepared_data=prepared_data)
        self.log(message)
        if OK:
            self.logger.info('1 record sent')
            if timestamp:
                self.last_update = timestamp
                self.context.status.set(
                    'last update', self.service_name, str(timestamp))
        return OK


class FileService(ServiceBase):
    """Base class for "file" services.

    """
    def do_catchup(self, do_all=False):
        self.upload(options=literal_eval(
            self.context.status.get('pending', self.service_name, '[]')))
        return True

    def upload(self, live_data=None, options=()):
        for item in options:
            if self.queue.full() or (item in self.queue):
                continue
            self.queue.append(item)

    def upload_batch(self):
        pending = literal_eval(
            self.context.status.get('pending', self.service_name, '[]'))
        OK = True
        count = 0
        with self.session() as session:
            while self.queue and not self.context.shutdown.is_set():
                upload = self.queue[0]
                if upload is None:
                    OK = False
                    break
                if os.path.isabs(upload):
                    path = upload
                else:
                    path = os.path.join(self.context.output_dir, upload)
                if not os.path.isfile(path):
                    if upload in pending:
                        pending.remove(upload)
                    self.queue.popleft()
                    continue
                self.logger.debug('file: %s', path)
                OK, message = self.upload_file(session, path)
                self.log(message)
                if OK:
                    if upload in pending:
                        pending.remove(upload)
                    count += 1
                else:
                    if upload not in pending:
                        pending.append(upload)
                    break
                self.queue.popleft()
        self.context.status.set('pending', self.service_name, repr(pending))
        if count > 1:
            self.logger.info('{:d} uploads'.format(count))
        elif count:
            self.logger.info('1 upload')
        return OK


def main(class_, argv=None):
    import argparse
    import inspect

    if argv is None:
        argv = sys.argv
    docstring = inspect.getdoc(sys.modules[class_.__module__])
    if sys.version_info[0] < 3:
        docstring = docstring.decode('utf-8')
    docstring = docstring.split('\n\n')
    parser = argparse.ArgumentParser(
        description=docstring[0], epilog=docstring[1])
    if hasattr(class_, 'register'):
        parser.add_argument('-r', '--register', action='store_true',
                            help='register (or update) with service')
    if issubclass(class_, CatchupDataService):
        parser.add_argument('-c', '--catchup', action='store_true',
                            help='upload all data since last upload')
    parser.add_argument('-v', '--verbose', action='count',
                        help='increase amount of reassuring messages')
    parser.add_argument('data_dir', help='root directory of the weather data')
    if issubclass(class_, FileService):
        parser.add_argument('file', nargs='*', help='file to be uploaded')
    args = parser.parse_args(argv[1:])
    pywws.logger.setup_handler(args.verbose or 0)
    with pywws.storage.pywws_context(args.data_dir) as context:
        if 'register' in args and args.register:
            uploader = class_(context, check_params=False)
            uploader.register()
            context.flush()
            return 0
        uploader = class_(context)
        if issubclass(class_, FileService):
            uploader.upload(options=map(os.path.abspath, args.file))
        elif issubclass(class_, CatchupDataService) and args.catchup:
            uploader.do_catchup(do_all=True)
        else:
            uploader.upload(test_mode=True)
        uploader.stop()
    return 0



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2018-20  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Upload weather data to MQTT message broker.

MQTT_ is a "message broker" system, typically running on ``localhost``
or another computer in your home network. Use of MQTT with pywws
requires an additional library. See :ref:`Dependencies - MQTT
<dependencies-mqtt>` for details.

* Mosquitto (a lightweight broker): http://mosquitto.org/
* Example ``weather.ini`` configuration::

    [mqtt]
    topic = /weather/pywws
    hostname = localhost
    port = 1883
    client_id = pywws
    retain = False
    user =
    password =
    tls_cert = /home/pi/pywws/ca_cert/mqtt_ca.crt
    tls_ver = 2
    multi_topic = False

    [logged]
    services = ['mqtt', 'underground']

* To customize the MQTT message use template_txt (remove illuminace and uv if weather station does not support them)::

    [mqtt]
    ... (as above)
    template_txt = ('\\n'
            '#idx          \\'"idx"         : "%Y-%m-%d %H:%M:%S",\\'#\\n'
            '#wind_dir     \\'"wind_dir_degrees"    : "%.d",\\' \\'\\' \\'winddir_degrees(x)\\'#\\n'
            '#wind_dir     \\'"wind_dir_text"       : "%s",\\' \\'\\' \\'winddir_text(x)\\'#\\n'
            '#wind_ave     \\'"wind_ave_mps"    : "%.2f",\\'#\\n'
            '#wind_ave     \\'"wind_ave_mph"    : "%.2f",\\' \\'\\' \\'wind_mph(x)\\'#\\n'
            '#wind_gust    \\'"wind_gust_mps"   : "%.2f",\\'#\\n'
            '#wind_gust    \\'"wind_gust_mph"   : "%.2f",\\' \\'\\' \\'wind_mph(x)\\'#\\n'
            '#calc \\'wind_chill(data["temp_out"],data["wind_ave"])\\' \\'"wind_chill_c" : "%.1f",\\'#\\n'
            '#calc \\'temp_f(wind_chill(data["temp_out"],data["wind_ave"]))\\' \\'"wind_chill_f" : "%.1f",\\'#\\n'
            '#calc \\'dew_point(data["temp_out"],data["hum_out"])\\' \\'"dew_point_c" : "%.1f",\\'#\\n'
            '#calc \\'temp_f(dew_point(data["temp_out"],data["hum_out"]))\\' \\'"dew_point_f" : "%.1f",\\'#\\n'
            '#hum_out      \\'"hum_out"     : "%.d",\\'#\\n'
            '#hum_in       \\'"hum_in"      : "%.d",\\'#\\n'
            '#temp_in      \\'"temp_in_c"   : "%.1f",\\'#\\n'
            '#temp_in      \\'"temp_in_f"   : "%.1f",\\' \\'\\' \\'temp_f(x)\\'#\\n'
            '#temp_out     \\'"temp_out_c"  : "%.1f",\\'#\\n'
            '#temp_out     \\'"temp_out_f"  : "%.1f",\\' \\'\\' \\'temp_f(x)\\'#\\n'
            '#calc \\'apparent_temp(data["temp_out"],data["hum_out"],data["wind_ave"])\\' \\'"temp_out_realfeel_c" : "%.1f",\\'#\\n'
            '#calc \\'temp_f(apparent_temp(data["temp_out"],data["hum_out"],data["wind_ave"]))\\' \\'"temp_out_realfeel_f" : "%.1f",\\'#\\n'
            '#rel_pressure \\'"pressure_rel_hpa": "%.1f",\\'#\\n'
            '#rel_pressure \\'"pressure_rel_inhg": "%.4f",\\' \\'\\' \\'pressure_inhg(x)\\'#\\n'
            '#abs_pressure \\'"pressure_abs_hpa": "%.1f",\\'#\\n'
            '#abs_pressure \\'"pressure_abs_inhg": "%.4f",\\' \\'\\' \\'pressure_inhg(x)\\'#\\n'
            '#rain         \\'"rain_mm"     : "%.1f",\\'#\\n'
            '#rain         \\'"rain_in"     : "%.2f",\\' \\'\\' \\'rain_inch(x)\\'#\\n'
            '#calc \\'rain_hour(data)\\' \\'"rain_last_hour_mm": "%.1f",\\'#\\n'
            '#calc \\'rain_inch(rain_hour(data))\\' \\'"rain_last_hour_in": "%.2f",\\'#\\n'
            '#calc \\'rain_24hr(data)\\' \\'"rain_last_24hours_mm": "%.1f",\\'#\\n'
            '#calc \\'rain_inch(rain_24hr(data))\\' \\'"rain_last_24hours_in": "%.2f",\\'#\\n'
            '#calc \\'rain_day(data)\\' \\'"rain_day_mm": "%.1f",\\'#\\n'
            '#calc \\'rain_inch(rain_day(data))\\' \\'"rain_day_in": "%.2f",\\'#\\n'
            '#illuminance  \\'"illuminance_lux" : "%.1f",\\'#\\n'
            '#illuminance  \\'"illuminance_wm2" : "%.2f",\\' \\'\\' \\'illuminance_wm2(x)\\'#\\n'
            '#uv           \\'"uv"          : "%.d",\\'#\\n'
            '\\n')

pywws will publish a JSON string of weather data. This data will be
published to the broker running on ``hostname``, with the port number
specified. (An IP address can be used instead of a host name.)
``client_id`` is a note of who published the data to the topic.
``topic`` can be any string value, this needs to be the topic that a
subscriber is aware of.

``retain`` is a boolean and should be set to ``True`` or ``False``. If
set to ``True`` this will flag the message sent to the broker to be
retained. Otherwise the broker discards the message if no client is
subscribing to this topic. This allows clients to get an immediate
response when they subscribe to a topic, without having to wait until
the next message is published.

``user`` and ``password`` can be used for MQTT authentication.

``tls_cert`` and ``tls_ver`` are used for MQTT TLS security. Set
tls_cert as the path to a CA certificate (e.g. tls_cert =
/home/pi/pywws/ca_cert/mqtt_ca.crt) and tls_ver to the TLS version (e.g.
tls_ver = 2) (TLS1.2 recommended). See
https://mosquitto.org/man/mosquitto-tls-7.html for information on how to
generate certificates. Only copy the ca.crt to your pywws client. See
http://www.steves-internet-guide.com/mosquitto-tls/ for a step-by-step
guide to securing your MQTT server. Note that secure MQTTS usually uses
port 8883, so you will need to also change the port number.

``multi_topic`` is a boolean and should be set to ``True`` or ``False``.
If set to ``True`` pywws will also publish all the data each as separate
subtopics of the configured ``topic``; e.g., with the ``topic`` set to
/weather/pywws pywws will also publish the outside temperature to
``/weather/pywws/temp_out_c`` and the inside temperature to
``/weather/pywws/temp_in_c``.

``template_txt`` is the template used to generate the data to be
published. You can edit it to suit your own requirements. Be very careful
about the backslash escaped quotation marks though. If not specified
default will be used, which sends a lot of values in metric and imperial
units.

.. versionchanged:: 19.5.0
   Default for ``template_txt`` was updated. This change is not backwards
   compatible, the original values are still present, just under new names.
   New default tries to send most of the values pywws collects in both
   metric and imperial units. This is to make it easier for new users
   to get going.

If these aren't obvious to you it's worth doing a bit of reading around
MQTT. It's a great lightweight messaging system from IBM, recently made
more popular when Facebook published information on their use of it.

This has been tested with the Mosquitto Open Source MQTT broker, running
on a Raspberry Pi (Raspian OS). TLS (mqtt data encryption) is not yet
implemented.

Thanks to Matt Thompson for writing the MQTT code and to Robin Kearney
for adding the retain and auth options.

.. _MQTT: http://mqtt.org/

"""

from __future__ import absolute_import, unicode_literals

from ast import literal_eval
from contextlib import contextmanager
from datetime import timedelta
import json
import logging
import os
import pprint
import sys

import paho.mqtt.client as mosquitto

import pywws.service

__docformat__ = "restructuredtext en"
service_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(__name__)


class ToService(pywws.service.LiveDataService):
    config = {
        'topic'      : ('/weather/pywws', True,  None),
        'hostname'   : ('localhost',      True,  None),
        'port'       : ('1883',           True,  None),
        'client_id'  : ('pywws',          True,  None),
        'retain'     : ('False',          True,  None),
        'user'       : ('',               False, None),
        'password'   : ('',               False, None),
        'tls_cert'   : ('',               False, None),
        'tls_ver'    : ('1',              True,  None),
        'multi_topic': ('False',          True,  None),
        }
    logger = logger
    service_name = service_name
    template = """
#idx          '"idx"             : "%Y-%m-%d %H:%M:%S",'#
#wind_dir     '"wind_dir_degrees": "%.d",' '' 'winddir_degrees(x)'#
#wind_dir     '"wind_dir_text"   : "%s",' '' 'winddir_text(x)'#
#wind_ave     '"wind_ave_mps"    : "%.2f",'#
#wind_ave     '"wind_ave_mph"    : "%.2f",' '' 'wind_mph(x)'#
#wind_gust    '"wind_gust_mps"   : "%.2f",'#
#wind_gust    '"wind_gust_mph"   : "%.2f",' '' 'wind_mph(x)'#
#calc 'wind_chill(data["temp_out"],data["wind_ave"])'         '"wind_chill_c" : "%.1f",'#
#calc 'temp_f(wind_chill(data["temp_out"],data["wind_ave"]))' '"wind_chill_f" : "%.1f",'#
#calc 'dew_point(data["temp_out"],data["hum_out"])'           '"dew_point_c" : "%.1f",'#
#calc 'temp_f(dew_point(data["temp_out"],data["hum_out"]))'   '"dew_point_f" : "%.1f",'#
#hum_out      '"hum_out"     : "%.d",'#
#hum_in       '"hum_in"      : "%.d",'#
#temp_in      '"temp_in_c"   : "%.1f",'#
#temp_in      '"temp_in_f"   : "%.1f",' '' 'temp_f(x)'#
#temp_out     '"temp_out_c"  : "%.1f",'#
#temp_out     '"temp_out_f"  : "%.1f",' '' 'temp_f(x)'#
#calc 'apparent_temp(data["temp_out"],data["hum_out"],data["wind_ave"])'         '"temp_out_realfeel_c" : "%.1f",'#
#calc 'temp_f(apparent_temp(data["temp_out"],data["hum_out"],data["wind_ave"]))' '"temp_out_realfeel_f" : "%.1f",'#
#rel_pressure '"pressure_rel_hpa" : "%.1f",'#
#rel_pressure '"pressure_rel_inhg": "%.4f",' '' 'pressure_inhg(x)'#
#abs_pressure '"pressure_abs_hpa" : "%.1f",'#
#abs_pressure '"pressure_abs_inhg": "%.4f",' '' 'pressure_inhg(x)'#
#rain         '"rain_mm"     : "%.1f",'#
#rain         '"rain_in"     : "%.2f",' '' 'rain_inch(x)'#
#calc 'rain_hour(data)'            '"rain_last_hour_mm": "%.1f",'#
#calc 'rain_inch(rain_hour(data))' '"rain_last_hour_in": "%.2f",'#
#calc 'rain_24hr(data)'            '"rain_last_24hours_mm": "%.1f",'#
#calc 'rain_inch(rain_24hr(data))' '"rain_last_24hours_in": "%.2f",'#
#calc 'rain_day(data)'             '"rain_day_mm": "%.1f",'#
#calc 'rain_inch(rain_day(data))'  '"rain_day_in": "%.2f",'#
"""
    template_3080_add = """
#illuminance  '"illuminance_lux" : "%.1f",'#
#illuminance  '"illuminance_wm2" : "%.2f",' '' 'illuminance_wm2(x)'#
#uv           '"uv"          : "%.d",'#
"""

    def __init__(self, context, check_params=True):
        super(ToService, self).__init__(context, check_params)
        # extend template
        if context.params.get('config', 'ws type') == '3080':
            self.template += self.template_3080_add
        # get template text
        template = literal_eval(context.params.get(
            service_name, 'template_txt', self.template_format(self.template)))
        logger.log(logging.DEBUG - 1, 'template:\n' + template)
        self.template = "#live#" + template
        # convert some params from string
        for key in ('port', 'retain', 'tls_ver', 'multi_topic'):
            self.params[key] = literal_eval(self.params[key])

    def template_format(self, template):
        result = []
        for line in template.splitlines():
            if line:
                result.append(pprint.pformat(line, width=256))
        return '(\n' + '\n'.join(result) + '\n)'

    @contextmanager
    def session(self):
        session = mosquitto.Client(
            self.params['client_id'], protocol=mosquitto.MQTTv31)
        if self.params['password']:
            session.username_pw_set(
                self.params['user'], self.params['password'])
        elif self.params['user']:
            session.username_pw_set(self.params['user'])
        logger.debug(('connecting to host {hostname:s}:{port:d} '
                      'with client_id "{client_id:s}"').format(**self.params))
        if self.params['tls_cert']:
            session.tls_set(self.params['tls_cert'], tls_version=self.params['tls_ver'])
        session.connect(self.params['hostname'], self.params['port'])
        try:
            yield session
        finally:
            session.disconnect()

    def upload_data(self, session, prepared_data={}):
        logger.debug((
            'publishing on topic "{topic:s}" with retain={retain!s},'
            ' data="{data!r}"').format(data=prepared_data, **self.params))
        try:
            session.publish(self.params['topic'], json.dumps(prepared_data),
                            retain=self.params['retain'])
        except Exception as ex:
            return False, repr(ex)
        if self.params['multi_topic']:
            # Publish messages, one for each item in prepared_data to
            # separate Subtopics.
            for key, value in prepared_data.items():
                if value == '':
                    value = 'None'
                try:
                    session.publish(self.params['topic'] + "/" + key, value,
                                    retain=self.params['retain'])
                except Exception as ex:
                    return False, repr(ex)
            # Need to make sure the messages have been flushed to the
            # server.
            session.loop(timeout=0.5)
        return True, 'OK'


if __name__ == "__main__":
    sys.exit(pywws.service.main(ToService))



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-21  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Routines to perform common tasks such as plotting gaphs or uploading files."""

from __future__ import absolute_import

from ast import literal_eval
from datetime import datetime, timedelta
import importlib
import logging
import os
import sys
import time

from pywws.calib import Calib
import pywws.plot
import pywws.template
from pywws.timezone import time_zone
import pywws.windrose

logger = logging.getLogger(__name__)


class RegularTasks(object):
    def __init__(self, context):
        self.context = context
        self.params = context.params
        self.status = context.status
        self.raw_data = context.raw_data
        self.calib_data = context.calib_data
        self.hourly_data = context.hourly_data
        self.daily_data = context.daily_data
        self.monthly_data = context.monthly_data
        self.flush = literal_eval(
            self.params.get('config', 'frequent writes', 'False'))
        # get directories
        self.template_dir = self.params.get(
            'paths', 'templates', os.path.expanduser('~/weather/templates/'))
        self.graph_template_dir = self.params.get(
            'paths', 'graph_templates',
            os.path.expanduser('~/weather/graph_templates/'))
        self.module_dir = self.params.get(
            'paths', 'modules', os.path.expanduser('~/weather/modules/'))
        # create calibration object
        self.calibrator = Calib(self.params, self.raw_data)
        # create templater object
        self.templater = pywws.template.Template(context)
        # create plotter objects
        self.plotter = pywws.plot.GraphPlotter(context, self.context.work_dir)
        self.roseplotter = pywws.windrose.RosePlotter(
            context, self.context.work_dir)
        # get daytime end hour
        self.day_end_hour, self.use_dst = pywws.process.get_day_end_hour(
                                                                self.params)
        # parse "cron" sections
        self.cron = {}
        for section in self.params._config.sections():
            if section.split()[0] != 'cron':
                continue
            import croniter
            last_update = self.status.get_datetime('last update', section)
            last_update = last_update or datetime.utcnow()
            self.cron[section] = croniter.croniter(
                self.params.get(section, 'format', ''),
                start_time=time_zone.utc_to_local(last_update))
            self.cron[section].get_next()
        # create service uploader objects
        self.services = {}
        for section in list(self.cron.keys()) + [
                       'live', 'logged', 'hourly', '12 hourly', 'daily']:
            for name, options in self._parse_templates(section, 'services'):
                if name in self.services:
                    continue
                if os.path.exists(os.path.join(self.module_dir, name + '.py')):
                    sys.path.insert(0, self.module_dir)
                    mod = importlib.import_module(name)
                    del sys.path[0]
                else:
                    mod = importlib.import_module('pywws.service.' + name)
                self.services[name] = mod.ToService(context)
            # check for obsolete entries
            if self.params.get(section, 'twitter'):
                logger.error(
                    'Obsolete twitter entry in weather.ini [%s]', section)
            if self.params.get(section, 'yowindow'):
                logger.error(
                    'Obsolete yowindow entry in weather.ini [%s]', section)

    def has_live_tasks(self):
        if self.cron:
            return True
        for name in literal_eval(self.params.get('live', 'services', '[]')):
            return True
        for template in literal_eval(self.params.get('live', 'plot', '[]')):
            return True
        for template in literal_eval(self.params.get('live', 'text', '[]')):
            return True
        return False

    def _parse_templates(self, section, option):
        for template in literal_eval(self.params.get(section, option, '[]')):
            if isinstance(template, (list, tuple)):
                yield template[0], template[1:]
            else:
                yield template, ()

    def _do_common(self, now, sections, live_data=None):
        logger.info('doing task sections {!r}'.format(sections))
        # make lists of tasks from all sections, avoiding repeats
        service_tasks = []
        text_tasks = []
        plot_tasks = []
        for section in sections:
            for task in self._parse_templates(section, 'services'):
                if task not in service_tasks:
                    service_tasks.append(task)
            for task in self._parse_templates(section, 'text'):
                if task not in text_tasks:
                    text_tasks.append(task)
            for task in self._parse_templates(section, 'plot'):
                if task not in plot_tasks:
                    plot_tasks.append(task)
        # do plot templates
        for template, flags in plot_tasks:
            self.do_plot(template)
        # do text templates
        for template, flags in text_tasks:
            self.do_template(template, data=live_data)
        # do service tasks
        for name, options in service_tasks:
            self.services[name].upload(live_data=live_data, options=options)
        # allow all services to sent some catchup records
        catchup = list(self.services.keys())
        stop = time.time() + 20.0
        while catchup and time.time() < stop:
            for name in list(catchup):
                if self.services[name].do_catchup():
                    catchup.remove(name)
        # update status
        for section in sections:
            self.status.set('last update', section, now.isoformat(' '))
        # save any unsaved data
        if self.flush or 'hourly' in sections:
            self.context.flush()

    def _cron_due(self, now):
        if not self.cron:
            return []
        # make list of due sections
        sections = []
        for section in self.cron:
            if time_zone.local_to_utc(
                    self.cron[section].get_current(datetime)) > now:
                continue
            sections.append(section)
            while time_zone.local_to_utc(
                    self.cron[section].get_next(datetime)) <= now:
                pass
        return sections

    def _periodic_due(self, now):
        # make list of due sections
        sections = []
        # hourly
        threshold = time_zone.hour_start(now)
        last_update = self.status.get_datetime('last update', 'hourly')
        if not last_update or last_update < threshold:
            sections.append('hourly')
        # daily
        threshold = time_zone.day_start(
            now, self.day_end_hour, use_dst=self.use_dst)
        last_update = self.status.get_datetime('last update', 'daily')
        if not last_update or last_update < threshold:
            sections.append('daily')
        # 12 hourly
        threshold = max(threshold, time_zone.day_start(
            now, (self.day_end_hour + 12) % 24, use_dst=self.use_dst))
        last_update = self.status.get_datetime('last update', '12 hourly')
        if not last_update or last_update < threshold:
            sections.append('12 hourly')
        return sections

    def do_live(self, data):
        calib_data = self.calibrator.calib(data)
        now = calib_data['idx']
        sections = ['live'] + self._cron_due(now) + self._periodic_due(now)
        self._do_common(now, sections, live_data=calib_data)

    def do_tasks(self):
        now = self.calib_data.before(datetime.max)
        if not now:
            raise RuntimeError('No processed data available')
        if not self.context.live_logging:
            # do periodic tasks if they would be due by next logging time
            now += timedelta(minutes=self.calib_data[now]['delay'])
        sections = ['logged'] + self._cron_due(now) + self._periodic_due(now)
        self._do_common(now, sections)
        if not self.context.live_logging:
            # cleanly shut down upload threads
            for name in self.services:
                self.services[name].stop()

    def do_plot(self, template):
        logger.info("Graphing %s", template)
        input_file = os.path.join(self.graph_template_dir, template)
        output_file = os.path.join(
            self.context.output_dir, os.path.splitext(template)[0])
        input_xml = pywws.plot.GraphFileReader(input_file)
        if (input_xml.get_children(self.plotter.plot_name) and
                        self.plotter.do_plot(input_xml, output_file) == 0):
            return output_file
        if (input_xml.get_children(self.roseplotter.plot_name) and
                        self.roseplotter.do_plot(input_xml, output_file) == 0):
            return output_file
        logger.warning('nothing to graph in %s', input_file)
        return None

    def do_template(self, template, data=None):
        logger.info("Templating %s", template)
        input_file = os.path.join(self.template_dir, template)
        output_file = os.path.join(self.context.output_dir, template)
        self.templater.make_file(input_file, output_file, live_data=data)
        return output_file



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-18  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Upload files to a web server by FTP.

This module uploads files to (typically) a website *via* FTP. Details of
the upload destination are stored in the file ``weather.ini`` in your
data directory. You should be able to get the required information from
your web space provider. If your provider allows SFTP then you could use
:py:mod:`pywws.service.sftp` for greater security.

* Example ``weather.ini`` configuration::

    [ftp]
    site = ftp.xxxx.yyyy.co.uk
    user = xxxxxxx
    password = zzzzzzzzz
    directory = public_html/weather/data/
    port = 21

    [hourly]
    plot = ['24hrs.png.xml', 'rose_12hrs.png.xml']
    text = ['24hrs.txt']
    services = [('ftp', '24hrs.txt', '24hrs.png', 'rose_12hrs.png')]

Run :py:mod:`pywws.service.ftp` once to set the default configuration,
which you can then change. ``directory`` is the name of a directory in
which all the uploaded files will be put. This will depend on the
structure of your web site and the sort of host you use. ``port`` is the
port number to use. 21 is the standard value but your web space provider
may require a different port.

You can upload any files you like, as often as you like, but typical
usage is to update a website once an hour. Each file to be uploaded
needs to be listed in a service entry like ``('ftp', 'filename')``. If
the file is not in your ``work`` directory's ``output`` directory then
``filename`` should be the full path.

"""

from __future__ import absolute_import

from contextlib import closing, contextmanager
from datetime import timedelta
import ftplib
import logging
import os
import sys

import pywws.service

__docformat__ = "restructuredtext en"
service_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(__name__)


class ToService(pywws.service.FileService):
    config = {
        'site'       : ('',   True, None),
        'user'       : ('',   True, None),
        'password'   : ('',   True, None),
        'directory'  : ('',   True, None),
        'port'       : ('21', True, None),
        }
    logger = logger
    service_name = service_name

    def __init__(self, context, check_params=True):
        super(ToService, self).__init__(context, check_params)
        self.params['port'] = int(self.params['port'])

    @contextmanager
    def session(self):
        with closing(ftplib.FTP()) as session:
            session.connect(self.params['site'], self.params['port'])
            logger.debug('welcome message\n' + session.getwelcome())
            session.login(self.params['user'], self.params['password'])
            session.cwd(self.params['directory'])
            yield session

    def upload_file(self, session, path):
        target = os.path.basename(path)
        text_file = os.path.splitext(target)[1] in ('.txt', '.xml', '.html')
        if text_file and sys.version_info[0] < 3:
            mode = 'r'
        else:
            mode = 'rb'
        try:
            with open(path, mode) as f:
                if text_file:
                    session.storlines('STOR %s' % (target), f)
                else:
                    session.storbinary('STOR %s' % (target), f)
        except Exception as ex:
            return False, repr(ex)
        return True, 'OK'


if __name__ == "__main__":
    sys.exit(pywws.service.main(ToService))



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-20  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Set some weather station parameters.

This script can also be run with the ``pywws-setweatherstation`` command. ::
%s

"""

from __future__ import absolute_import, print_function

__docformat__ = "restructuredtext en"
__usage__ = """
 usage: python -m pywws.setweatherstation [options]
 options are:
  -h   | --help           display this help
  -c   | --clock          set weather station clock to computer time
                          (unlikely to work)
  -d   | --default        set sensible default values for some settings
  -p f | --pressure f     set relative pressure to f hPa
  -r n | --read_period n  set logging interval to n minutes
  -v   | --verbose        increase error message verbosity
  -z   | --zero_memory    clear the weather station logged reading count
"""
__doc__ %= __usage__
__usage__ = __doc__.split('\n')[0] + __usage__


from datetime import datetime, timedelta
import getopt
import logging
import sys
import time

import pywws.logger
import pywws.weatherstation


def bcd_encode(value):
    hi = value // 10
    lo = value % 10
    return (hi * 16) + lo


def main(argv=None):
    if argv is None:
        argv = sys.argv
    try:
        opts, args = getopt.getopt(
            argv[1:], "hcdp:r:vz",
            ['help', 'clock', 'default', 'pressure=', 'read_period=',
             'verbose', 'zero_memory'])
    except getopt.error as msg:
        print('Error: %s\n' % msg, file=sys.stderr)
        print(__usage__.strip(), file=sys.stderr)
        return 1
    # process options
    clock = False
    default = False
    pressure = None
    read_period = None
    verbose = 0
    zero_memory = False
    for o, a in opts:
        if o in ('-h', '--help'):
            print(__usage__.strip())
            return 0
        elif o in ('-c', '--clock'):
            clock = True
        elif o in ('-d', '--default'):
            default = True
        elif o in ('-p', '--pressure'):
            pressure = int((float(a) * 10.0) + 0.5)
        elif o in ('-r', '--read_period'):
            read_period = int(a)
        elif o in ('-v', '--verbose'):
            verbose += 1
        elif o in ('-z', '--zero_memory'):
            zero_memory = True
    # check arguments
    if len(args) != 0:
        print("Error: No arguments required", file=sys.stderr)
        print(__usage__.strip(), file=sys.stderr)
        return 2
    pywws.logger.setup_handler(verbose)
    # open connection to weather station
    ws = pywws.weatherstation.WeatherStation()
    # set data to be sent to station
    data = []
    # set default values
    if default:
        ptr = ws.fixed_format['settings_1'][0]
        data.append((ptr, 0b00100000))
        ptr = ws.fixed_format['settings_2'][0]
        data.append((ptr, 0b00001000))
        ptr = ws.fixed_format['display_1'][0]
        data.append((ptr, 0b01000001))
        ptr = ws.fixed_format['display_2'][0]
        data.append((ptr, 0b00001001))
        ptr = ws.fixed_format['alarm_1'][0]
        data.append((ptr, 0b00000000))
        ptr = ws.fixed_format['alarm_2'][0]
        data.append((ptr, 0b00000000))
        ptr = ws.fixed_format['alarm_3'][0]
        data.append((ptr, 0b00000000))
    # set relative pressure
    if pressure:
        ptr = ws.fixed_format['rel_pressure'][0]
        data.append((ptr,   pressure % 256))
        data.append((ptr+1, pressure // 256))
    # set read period
    if read_period:
        data.append((ws.fixed_format['read_period'][0], read_period))
    # reset data count
    if zero_memory:
        ptr = ws.fixed_format['data_count'][0]
        data.append((ptr,   1))
        data.append((ptr+1, 0))
    # set clock
    if clock:
        print("Clock setting is not known to work on any model of weather station.")
        print("If it works for you, please let Jim Easterbrook know.")
        print("waiting for exact minute")
        now = datetime.now()
        if now.second >= 55:
            time.sleep(10)
            now = datetime.now()
        now += timedelta(minutes=1)
        ptr = ws.fixed_format['date_time'][0]
        data.append((ptr,   bcd_encode(now.year - 2000)))
        data.append((ptr+1, bcd_encode(now.month)))
        data.append((ptr+2, bcd_encode(now.day)))
        data.append((ptr+3, bcd_encode(now.hour)))
        data.append((ptr+4, bcd_encode(now.minute)))
        time.sleep(59 - now.second)
    # send it all in one go
    if data:
        ws.write_data(data)


if __name__ == "__main__":
    sys.exit(main())



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-18  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Post messages to Twitter.

Twitter_ is a popular micro-blogging service. Users send short messages
(up to 280 characters) called "tweets". This module sends tweets, with
optional image files, typically to report on weather conditions every
hour.

* Create account: https://twitter.com/
* Additional dependencies:

  * https://github.com/joestump/python-oauth2
  * https://github.com/bear/python-twitter *or*
    https://github.com/tweepy/tweepy
* Example ``weather.ini`` configuration::

    [twitter]
    secret = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    key = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    latitude = 51.501
    longitude = -0.142

    [hourly]
    plot = ['tweet.png.xml', '24hrs.png.xml', 'rose_12hrs.png.xml']
    text = ['tweet_media.txt', '24hrs.txt']
    services = [('twitter', 'tweet_media.txt'),
                ('ftp', '24hrs.png', 'rose_12hrs.png', '24hrs.txt')]

Create an account
-----------------

You could post weather updates to your ‘normal’ Twitter account, but I
think it’s better to have a separate account just for weather reports.
This could be useful to someone who lives in your area, but doesn’t want
to know what you had for breakfast.

The :py:mod:`pywws.service.twitter` module requires you to install
additional dependencies. There is a choice of Python Twitter library.
Use either ``python-twitter`` (preferred) or ``tweepy``::

    sudo pip install python-twitter oauth2

*or* ::

    sudo pip install tweepy oauth2

Authorise pywws
---------------

Before you can send tweets you need to authorise pywws to post to your
account. If you run pywws on a low power device such as a Raspberry Pi,
you may find it easier to run this authorisation step on another
computer, as long as it has the required dependencies installed. You can
use an empty 'data' directory -- a ``weather.ini`` file will be created
whose contents can be copied into your real ``weather.ini`` file using
any text editor.

Make sure no other pywws software is running, then run the module with
the ``-r`` option. (Remember to replace ``data_dir`` with your data
directory.) ::

    python -m pywws.service.twitter -r data_dir

This will open a web browser window (or give you a URL to copy to your
web browser) where you can log in to your Twitter account and authorise
pywws to post. If the login is successful the browser will display a 7
digit number which you then copy to pywws::

    jim@brains:~$ python3 -m pywws.service.twitter -r weather_data
    14:19:21:pywws.logger:pywws version 18.8.0, build 1566 (5054dc7)
    Please enter the PIN shown in your web browser: 2594624
    jim@brains:~$ 

The ``secret`` and ``key`` values stored in ``weather.ini`` give access
to your Twitter account and should be kept confidential.

Add location data (optional)
----------------------------

Edit your ``weather.ini`` file and add ``latitude`` and ``longitude``
entries to the ``[twitter]`` section. For example::

   [twitter]
   secret = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
   key = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
   latitude = 51.501
   longitude = -0.142

Create a template
-----------------

A tweet is a short text of up to 280 characters (fewer if you include
images). It's up to you what to put in your tweets but an example is
included to get your started. Copy the example template ``tweet.txt`` to
your template directory, then edit it to suit your preferences. You
should also check it uses the same text encoding as your other
templates.

Now generate a tweet file from your template, for example::

    python -m pywws.template ~/weather/data ~/weather/templates/tweet.txt tweet.txt
    cat tweet.txt

(Replace ``~/weather/data`` and ``~/weather/templates`` with your data
and template directories.) If you need to change the template (e.g. to
change the units or language used) you can edit it now or later.

Post your first tweet
---------------------

Now you are ready to run :py:mod:`pywws.service.twitter`. Using high
verbosity shows you what's happening as it runs::

    python -m pywws.service.twitter -vv ~/weather/data tweet.txt

If this works, your new Twitter account will have posted its first
weather report. (You can delete the tweet.txt file now.)

Add Twitter posts to your hourly tasks
--------------------------------------

Edit the ``[hourly]`` section in ``weather.ini``. Add your tweet
template to the ``text`` list. Then add ``twitter`` to the ``services``
list, with an option specifying the template processing result. For
example::

    [hourly]
    text = ['tweet.txt']
    services = [('twitter', 'tweet.txt')]

You could use the ``[logged]``, ``[12 hourly]`` or ``[daily]`` sections
instead, but I think ``[hourly]`` is most appropriate for Twitter
updates.

Include images in your tweet
----------------------------

You can add up to four images to your tweets by specifying the image
file locations in the tweet template. Make the first line of the tweet
``media path`` where ``path`` is the file name, or full path for files
that are not in your "output" directory (a subdirectory of your work
directory called ``output``). Repeat for any additional image files. The
"tweet_media.txt" example template shows how to do this.

The image could be from a web cam, or for a weather forecast it could be
an icon representing the forecast. To add a weather graph you need to
make sure the graph is drawn before the tweet is sent. The
:py:mod:`pywws.regulartasks` module processes graph and text templates
before doing service uploads, so you can include the graph drawing in
the same section. For example::

    [hourly]
    plot = ['tweet.png.xml']
    text = ['tweet_media.txt']
    services = [('twitter', 'tweet_media.txt')]

.. _Twitter: https://twitter.com/

"""

from __future__ import absolute_import, print_function

import codecs
from contextlib import contextmanager
import logging
import os
import sys

twitter = None
tweepy = None
try:
    import twitter
except ImportError as ex:
    try:
        import tweepy
    except ImportError:
        # raise exception on the preferred library
        raise ex

from pywws.constants import Twitter as pct
import pywws.service

__docformat__ = "restructuredtext en"
service_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(__name__)


class TweepyHandler(object):
    def __init__(self, key, secret, latitude, longitude):
        logger.info('Using tweepy library')
        auth = tweepy.OAuthHandler(pct.consumer_key, pct.consumer_secret)
        auth.set_access_token(key, secret)
        self.api = tweepy.API(auth)
        if latitude is not None and longitude is not None:
            self.kwargs = {'lat' : latitude, 'long' : longitude}
        else:
            self.kwargs = {}

    def post(self, status, media):
        if len(media) > 1:
            logger.error('Tweepy library cannot post multiple media')
        if media:
            self.api.update_with_media(media[0], status[:257], **self.kwargs)
        else:
            self.api.update_status(status[:280], **self.kwargs)


class PythonTwitterHandler(object):
    def __init__(self, key, secret, latitude, longitude, timeout):
        logger.info('Using python-twitter library')
        self.api = twitter.Api(
            consumer_key=pct.consumer_key,
            consumer_secret=pct.consumer_secret,
            access_token_key=key, access_token_secret=secret,
            timeout=timeout)
        if latitude is not None and longitude is not None:
            self.kwargs = {'latitude' : latitude, 'longitude' : longitude,
                           'display_coordinates' : True}
        else:
            self.kwargs = {}
        self.kwargs['verify_status_length'] = False


    def post(self, status, media):
        max_len = 280
        if media:
            max_len -= len(media[:4]) * 23
        status = status.strip()[:max_len]
        args = dict(self.kwargs)
        if media:
            args['media'] = media
        self.api.PostUpdate(status, **args)


class ToService(pywws.service.FileService):
    config = {
        'secret'   : (None, True,  None),
        'key'      : (None, True,  None),
        'latitude' : (None, False, None),
        'longitude': (None, False, None),
        }
    logger = logger
    service_name = service_name

    def __init__(self, context, check_params=True):
        super(ToService, self).__init__(context, check_params)
        # get default character encoding of template output
        self.encoding = context.params.get(
            'config', 'template encoding', 'iso-8859-1')

    @contextmanager
    def session(self):
        if twitter:
            yield PythonTwitterHandler(
                self.params['key'], self.params['secret'],
                self.params['latitude'], self.params['longitude'], 40)
        else:
            yield TweepyHandler(
                self.params['key'], self.params['secret'],
                self.params['latitude'], self.params['longitude'])

    def upload_file(self, session, filename):
        with codecs.open(filename, 'r', encoding=self.encoding) as tweet_file:
            tweet = tweet_file.read()
        media = []
        while tweet.startswith('media'):
            media_item, tweet = tweet.split('\n', 1)
            media_item = media_item.split()[1]
            if not os.path.isabs(media_item):
                media_item = os.path.join(self.context.output_dir, media_item)
            media.append(media_item)
        try:
            session.post(tweet, media)
        except Exception as ex:
            message = repr(ex)
            return 'is a duplicate' in message, message
        return True, 'OK'

    def register(self):
        import webbrowser
        if sys.version_info[0] >= 3:
            from urllib.parse import parse_qsl
        else:
            from urlparse import parse_qsl
        import oauth2 as oauth

        consumer = oauth.Consumer(pct.consumer_key, pct.consumer_secret)
        client = oauth.Client(consumer)
        # step 1 - obtain a request token
        resp, content = client.request(
            'https://api.twitter.com/oauth/request_token', 'POST')
        if resp['status'] != '200':
            print('Failed to get request token. [%s]' % resp['status'])
            return
        if isinstance(content, bytes):
            content = content.decode('utf-8')
        request_token = dict(parse_qsl(content))
        # step 2 - redirect the user
        redirect_url = 'https://api.twitter.com/oauth/authorize?oauth_token=%s' % (
            request_token['oauth_token'])
        if not webbrowser.open(redirect_url, new=2, autoraise=0):
            print('Please use a web browser to open the following URL')
            print(redirect_url)
        if sys.version_info[0] >= 3:
            input_ = input
        else:
            input_ = raw_input
        pin = input_('Please enter the PIN shown in your web browser: ')
        pin = pin.strip()
        # step 3 - convert the request token to an access token
        token = oauth.Token(
            request_token['oauth_token'], request_token['oauth_token_secret'])
        token.set_verifier(pin)
        client = oauth.Client(consumer, token)
        resp, content = client.request(
            'https://api.twitter.com/oauth/access_token', 'POST')
        if resp['status'] != '200':
            print('Failed to get access token. [%s]' % resp['status'])
            return
        if isinstance(content, bytes):
            content = content.decode('utf-8')
        access_token = dict(parse_qsl(content))
        self.context.params.set(
            service_name, 'key', access_token['oauth_token'])
        self.context.params.set(
            service_name, 'secret', access_token['oauth_token_secret'])


if __name__ == "__main__":
    sys.exit(pywws.service.main(ToService))



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-18  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Get weather data, process it, prepare graphs & text files and
upload to a web site.

Typically run every hour from cron. ::

%s

This script does little more than call other modules in sequence to
get data from the weather station, process it, plot some graphs,
generate some text files and upload the results to a web site.

For more information on using ``pywws.hourly``, see
:doc:`../guides/hourlylogging`.

"""

from __future__ import absolute_import, print_function

__docformat__ = "restructuredtext en"
__usage__ = """
 usage: %s [options] data_dir
 options are:
  -h or --help     display this help
  -v or --verbose  increase amount of reassuring messages
 data_dir is the root directory of the weather data (e.g. $(HOME)/weather/data)
"""
__doc__ %= __usage__ % ('python -m pywws.hourly')

import getopt
import os
import sys

import pywws.localisation
import pywws.logdata
import pywws.logger
import pywws.process
import pywws.regulartasks
import pywws.storage


def hourly(data_dir):
    with pywws.storage.pywws_context(data_dir) as context:
        # localise application
        pywws.localisation.set_application_language(context.params)
        # get weather station data
        pywws.logdata.DataLogger(context).log_data()
        # do the processing
        pywws.process.process_data(context)
        # do tasks
        pywws.regulartasks.RegularTasks(context).do_tasks()
    return 0


def main(argv=None):
    if argv is None:
        argv = sys.argv
    usage = (__usage__ % (argv[0])).strip()
    try:
        opts, args = getopt.getopt(argv[1:], "hv", ['help', 'verbose'])
    except getopt.error as msg:
        print('Error: %s\n' % msg, file=sys.stderr)
        print(usage, file=sys.stderr)
        return 1
    # process options
    verbose = 0
    for o, a in opts:
        if o == '-h' or o == '--help':
            print(__doc__.split('\n\n')[0])
            print(usage)
            return 0
        elif o == '-v' or o == '--verbose':
            verbose += 1
    # check arguments
    if len(args) != 1:
        print('Error: 1 argument required\n', file=sys.stderr)
        print(usage, file=sys.stderr)
        return 2
    pywws.logger.setup_handler(verbose)
    return hourly(args[0])


if __name__ == "__main__":
    sys.exit(main())



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-18  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Configure Python logging system"""

from __future__ import absolute_import

import logging
import logging.handlers
import sys

from pywws import __version__, _release, _commit

logger = logging.getLogger(__name__)


class SystemdFormatter(logging.Formatter):
    def format(self, record):
        level = min((68 - record.levelno) // 8, 7)
        return '<{:d}>{:s}'.format(
            level, super(SystemdFormatter, self).format(record))

    def formatException(self, exc_info):
        msg = super(SystemdFormatter, self).formatException(exc_info)
        return msg.replace('\n', '\\n')


class FilterURLLib3(object):
    def __init__(self, level):
        self.level = level

    def filter(self, record):
        if (record.name == 'urllib3.connectionpool' and
                record.levelno < self.level + 1):
            return 0
        return 1


def setup_handler(verbose, logfile=None):
    root_logger = logging.getLogger('')
    if logfile=='systemd':
        level = logging.ERROR - (verbose * 10)
        handler = logging.StreamHandler(stream=sys.stdout)
        formatter = SystemdFormatter('%(name)s:%(message)s')
    elif logfile:
        level = logging.ERROR - (verbose * 10)
        handler = logging.handlers.RotatingFileHandler(
            logfile, maxBytes=128*1024, backupCount=3)
        formatter = logging.Formatter(
            '%(asctime)s:%(name)s:%(message)s', '%Y-%m-%d %H:%M:%S')
    else:
        level = logging.WARNING - (verbose * 10)
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            '%(asctime)s:%(name)s:%(message)s', '%H:%M:%S')
    level = max(level, 1)
    root_logger.setLevel(level)
    handler.addFilter(FilterURLLib3(level))
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)
    logger.warning(
        'pywws version %s, build %s (%s)', __version__, _release, _commit)
    logger.info('Python version %s', sys.version)



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2019-20  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Upload data to Windy.

Windy_  is a Czech company providing interactive weather forecasting
services worldwide.

* Create account: https://stations.windy.com/
* API: https://community.windy.com/topic/8168/report-your-weather-station-data-to-windy
* Forum discussion: https://community.windy.com/topic/11014/pywws-great-python-software-app-to-send-data-to-windy
* Additional dependency: http://docs.python-requests.org/
* Example ``weather.ini`` configuration::

    [windy]
    api_key = very.long.string.indeed
    station_id =

    [logged]
    services = ['windy', 'underground']

    [live]
    services = ['windy', 'underground']

Note that you only need to specify your station ID if you have more than
one station defined for your API. Windy allows other data, such as the
latitude and longitude, to be included in the upload but it's better to
set these by editing your station at
https://stations.windy.com/stations.

.. _Windy: http://www.windy.com

"""

from contextlib import contextmanager
from datetime import timedelta
import logging
import os
import sys

import requests

import pywws.service

__docformat__ = "restructuredtext en"
service_name = os.path.splitext(os.path.basename(__file__))[0]
logger = logging.getLogger(__name__)


class ToService(pywws.service.CatchupDataService):
    config = {
        'api_key'     : ('', True,  None),
        'station_id'  : ('', False, 'station'),
        }
    interval = timedelta(seconds=290)
    logger = logger
    service_name = service_name
    template = """
#live#
#idx          "'dateutc'     : '%Y-%m-%d %H:%M:%S',"#
#wind_ave     "'wind'        : '%.1f',"#
#wind_dir     "'winddir'     : '%.0f'," "" "winddir_degrees(x)"#
#wind_gust    "'gust'        : '%.2f',"#
#hum_out      "'humidity'    : '%.d',"#
#temp_out     "'temp'        : '%.1f',"#
#rel_pressure "'mbar'        : '%.1f',"#
#calc "dew_point(data['temp_out'], data['hum_out'])" "'dewpoint': '%.1f',"#
#calc "rain_hour(data)" "'precip': '%.2f',"#
"""

    def __init__(self, context, check_params=True):
        super(ToService, self).__init__(context, check_params)
        # extend template
        if context.params.get('config', 'ws type') == '3080':
            self.template += """
#uv           "'uv'            : '%d',"#
"""

    @contextmanager
    def session(self):
        with requests.Session() as session:
            yield session

    def upload_data(self, session, prepared_data={}):
        url = 'https://stations.windy.com/pws/update/' + self.params['api_key']
        try:
            rsp = session.get(url, params=prepared_data, timeout=60)
        except Exception as ex:
            return False, repr(ex)
        if rsp.status_code != 200:
            return False, 'http status: {:d}'.format(rsp.status_code)
        text = rsp.text.strip()
        return 'SUCCESS' in text, 'server response "{:s}"'.format(text)


if __name__ == "__main__":
    sys.exit(pywws.service.main(ToService))



# pywws - Python software for USB Wireless Weather Stations
# http://github.com/jim-easterbrook/pywws
# Copyright (C) 2008-19  pywws contributors

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Store weather data in easy to access files

Introduction
------------

This module is at the core of pywws file based storage.
It stores data on disc, but without the overhead of a full scale
database system. I have designed it to run on a small memory machine
such as a Raspberry Pi or even a router. To minimise memory usage it
only loads one day's worth of raw data at a time into memory.

From a "user" point of view, the data is accessed as a cross between a
list and a dictionary. Each data record is indexed by a
:py:class:`datetime.datetime` object (dictionary behaviour), but
records are stored in order and can be accessed as slices (list
behaviour).

For example, to access the hourly data for Christmas day 2009, one
might do the following::

  from datetime import datetime
  import pywws.filedata
  hourly = pywws.filedata.HourlyStore('weather_data')
  for data in hourly[datetime(2009, 12, 25):datetime(2009, 12, 26)]:
      print(data['idx'], data['temp_out'])

Some more examples of data access::

  # get value nearest 9:30 on Christmas day 2008
  data[data.nearest(datetime(2008, 12, 25, 9, 30))]
  # get entire array, equivalent to data[:]
  data[datetime.min:datetime.max]
  # get last 12 hours worth of data
  data[datetime.utcnow() - timedelta(hours=12):]

Note that the :py:class:`datetime.datetime` index is in UTC. You may
need to apply an offset to convert to local time.

The module provides five classes to store different data.
:py:class:`RawStore` takes "raw" data from the weather station;
:py:class:`CalibStore`, :py:class:`HourlyStore`, :py:class:`DailyStore`
and :py:class:`MonthlyStore` store processed data (see
:py:mod:`pywws.process`). All are derived from the same ``CoreStore``
class, they only differ in the keys and types of data stored in each
record.

Detailed API
------------

"""

import csv
from datetime import date, datetime, timedelta, MAXYEAR
import logging
import os
import sys
import time

from pywws.constants import DAY
from pywws.weatherstation import WSDateTime, WSFloat, WSInt, WSStatus

logger = logging.getLogger(__name__)


class _Cache(object):
    def __init__(self):
        self.data = []
        self.path = ''
        self.lo = date.max
        self.hi = date.min
        self.dirty = False

    def copy(self, other):
        self.data = other.data
        self.path = other.path
        self.lo = other.lo
        self.hi = other.hi
        self.dirty = False

    def get_ptr(self, idx):
        """Return index at which to insert a record with timestamp idx
        and boolean indicating if there is already data with that
        timestamp.

        """
        hi = len(self.data) - 1
        if hi < 0 or self.data[0]['idx'] > idx:
            return 0, False
        if self.data[0]['idx'] == idx:
            return 0, True
        if self.data[hi]['idx'] < idx:
            return hi + 1, False
        lo = 0
        while hi - lo > 1:
            mid = (lo + hi) // 2
            if self.data[mid]['idx'] < idx:
                lo = mid
            elif self.data[mid]['idx'] == idx:
                return mid, True
            else:
                hi = mid
        return hi, self.data[hi]['idx'] == idx


class CoreStore(object):
    def __init__(self, root_dir):
        self._root_dir = os.path.abspath(os.path.join(root_dir, self.dir_name))
        if not os.path.isdir(self._root_dir):
            os.mkdir(self._root_dir)
        # initialise caches
        self._wr_cache = _Cache()
        self._rd_cache = _Cache()
        # get conservative first and last days for which data (might) exist
        self._lo_limit = date.max - timedelta(days=500)
        self._hi_limit = date.min + timedelta(days=500)
        for root, dirs, files in os.walk(self._root_dir):
            files.sort()
            for file in files:
                if file[0] == '.':
                    continue
                path, self._lo_limit, hi = self._get_cache_path(
                    datetime.strptime(file, "%Y-%m-%d.txt").date())
                break
            else:
                dirs.sort()
                continue
            break
        for root, dirs, files in os.walk(self._root_dir):
            files.sort()
            files.reverse()
            for file in files:
                if file[0] == '.':
                    continue
                path, lo, self._hi_limit = self._get_cache_path(
                    datetime.strptime(file, "%Y-%m-%d.txt").date())
                break
            else:
                dirs.sort()
                dirs.reverse()
                continue
            break
        self._lo_limit_dt = datetime(
            self._lo_limit.year, self._lo_limit.month, self._lo_limit.day)
        self._hi_limit_dt = datetime(
            self._hi_limit.year, self._hi_limit.month, self._hi_limit.day)

    def _slice(self, i):
        if i.step is not None:
            raise TypeError("slice step not permitted")
        a = i.start
        if a is None:
            a = self._lo_limit_dt
        elif not isinstance(a, datetime):
            raise TypeError("slice indices must be %s or None" % (datetime))
        elif a < self._lo_limit_dt:
            a = self._lo_limit_dt
        b = i.stop
        if b is None:
            b = self._hi_limit_dt
        elif not isinstance(b, datetime):
            raise TypeError("slice indices must be %s or None" % (datetime))
        elif b > self._hi_limit_dt:
            b = self._hi_limit_dt
        return a, b

    def _get_slice(self, i):
        a, b = self._slice(i)
        if a > b:
            return
        # use separate cache as something might change self._rd_cache
        # during yield
        cache = _Cache()
        # go to start of slice
        start, exact = self._get_cache_ptr(cache, a)
        # iterate over complete caches
        while cache.hi <= b.date():
            for data in cache.data[start:]:
                yield data
            if cache.hi >= self._hi_limit:
                return
            self._load(cache, cache.hi)
            start = 0
        # iterate over part of cache
        for data in cache.data[start:]:
            if data['idx'] >= b:
                return
            yield data

    def __getitem__(self, i):
        """Return the data item or items with index i.

        i must be a datetime object or a slice.
        If i is a single datetime then a value with that index must exist."""
        if isinstance(i, slice):
            return self._get_slice(i)
        if not isinstance(i, datetime):
            raise TypeError("list indices must be %s" % (datetime))
        cache = self._rd_cache
        ptr, exact = self._get_cache_ptr(cache, i)
        if not exact:
            raise KeyError(i)
        return cache.data[ptr]

    def __setitem__(self, i, x):
        """Store a value x with index i.

        i must be a datetime object.
        If there is already a value with index i, it is overwritten.
        """
        if not isinstance(i, datetime):
            raise TypeError("index '%s' is not %s" % (i, datetime))
        x['idx'] = i
        cache = self._wr_cache
        ptr, exact = self._get_cache_ptr(cache, i)
        if exact:
            cache.data[ptr] = x
        else:
            cache.data.insert(ptr, x)
        cache.dirty = True

    def _del_slice(self, i):
        a, b = self._slice(i)
        if a > b:
            return
        # go to start of slice
        cache = self._wr_cache
        start, exact = self._get_cache_ptr(cache, a)
        # delete to end of cache
        while cache.hi <= b.date():
            del cache.data[start:]
            cache.dirty = True
            if cache.hi >= self._hi_limit:
                return
            self._load(cache, cache.hi)
            start = 0
        # delete part of cache
        stop, exact = cache.get_ptr(b)
        del cache.data[start:stop]
        cache.dirty = True

    def __delitem__(self, i):
        """Delete the data item or items with index i.

        i must be a datetime object or a slice.
        If i is a single datetime then a value with that index must exist."""
        if isinstance(i, slice):
            return self._del_slice(i)
        if not isinstance(i, datetime):
            raise TypeError("list indices must be %s" % (datetime))
        cache = self._wr_cache
        ptr, exact = self._get_cache_ptr(cache, i)
        if not exact:
            raise KeyError(i)
        del cache.data[ptr]
        cache.dirty = True

    def before(self, idx):
        """Return datetime of newest existing data record whose
        datetime is < idx.

        Might not even be in the same year! If no such record exists,
        return None."""
        if not isinstance(idx, datetime):
            raise TypeError("'%s' is not %s" % (idx, datetime))
        day = min(idx.date(), self._hi_limit - DAY)
        cache = self._rd_cache
        while day >= self._lo_limit:
            if day < cache.lo or day >= cache.hi:
                self._load(cache, day)
            ptr, exact = cache.get_ptr(idx)
            if ptr > 0:
                return cache.data[ptr - 1]['idx']
            day = cache.lo - DAY
        return None

    def after(self, idx):
        """Return datetime of oldest existing data record whose
        datetime is >= idx.

        Might not even be in the same year! If no such record exists,
        return None."""
        if not isinstance(idx, datetime):
            raise TypeError("'%s' is not %s" % (idx, datetime))
        day = max(idx.date(), self._lo_limit)
        cache = self._rd_cache
        while day < self._hi_limit:
            if day < cache.lo or day >= cache.hi:
                self._load(cache, day)
            ptr, exact = cache.get_ptr(idx)
            if ptr < len(cache.data):
                return cache.data[ptr]['idx']
            day = cache.hi
        return None

    def nearest(self, idx):
        """Return datetime of record whose datetime is nearest idx."""
        hi = self.after(idx)
        if hi == idx:
            return hi
        lo = self.before(idx)
        if hi is None:
            return lo
        if lo is None:
            return hi
        if (hi - idx) < (idx - lo):
            return hi
        return lo

    def _get_cache_ptr(self, cache, i):
        day = i.date()
        if day < cache.lo or day >= cache.hi:
            self._load(cache, day)
            if cache.lo < self._lo_limit:
                self._lo_limit = cache.lo
                self._lo_limit_dt = datetime(
                    cache.lo.year, cache.lo.month, cache.lo.day)
            if cache.hi > self._hi_limit:
                self._hi_limit = cache.hi
                self._hi_limit_dt = datetime(
                    cache.hi.year, cache.hi.month, cache.hi.day)
        return cache.get_ptr(i)

    def _load(self, cache, target_date):
        self._flush(cache)
        new_path, new_lo, new_hi = self._get_cache_path(target_date)
        if new_path == self._wr_cache.path:
            cache.copy(self._wr_cache)
            return
        if new_path == self._rd_cache.path:
            cache.copy(self._rd_cache)
            return
        cache.data = []
        cache.path, cache.lo, cache.hi = new_path, new_lo, new_hi
        if not os.path.exists(cache.path):
            return
        if sys.version_info[0] >= 3:
            kwds = {'mode': 'r', 'newline': ''}
        else:
            kwds = {'mode': 'rb'}
        row_lengths = (len(self.key_list),
                       len(self.key_list) - self.solar_items)
        with open(cache.path, **kwds) as csvfile:
            reader = csv.reader(csvfile, quoting=csv.QUOTE_NONE)
            for row in reader:
                if len(row) not in row_lengths:
                    logger.error('Invalid %s data at %s', self.dir_name, row[0])
                    continue
                result = {}
                for key, value in zip(self.key_list, row):
                    if value == '':
                        result[key] = None
                    else:
                        result[key] = self.conv[key](value)
                cache.data.append(result)

    def flush(self):
        self._flush(self._wr_cache)
        self._flush(self._rd_cache)

    def _flush(self, cache):
        if not cache.dirty:
            return
        cache.dirty = False
        if len(cache.data) == 0:
            if os.path.exists(cache.path):
                # existing data has been wiped, so delete file
                os.unlink(cache.path)
            return
        dir = os.path.dirname(cache.path)
        if not os.path.isdir(dir):
            os.makedirs(dir)
        if sys.version_info[0] >= 3:
            kwds = {'mode': 'w', 'newline': ''}
        else:
            kwds = {'mode': 'wb'}
        conv = {
            datetime  : str,
            float     : lambda x: '{:.12g}'.format(x),
            int       : str,
            type(None): lambda x: '',
            WSDateTime: WSDateTime.to_csv,
            WSFloat   : str,
            WSInt     : str,
            WSStatus  : WSStatus.to_csv,
            }
        with open(cache.path, **kwds) as csvfile:
            for data in cache.data:
                row = []
                for key in self.key_list[:len(data)]:
                    value = data[key]
                    row.append(conv[type(value)](value))
                csvfile.write(','.join(row) + '\n')

    def _get_cache_path(self, target_date):
        # default implementation - one file per day
        path = os.path.join(self._root_dir,
                            target_date.strftime("%Y"),
                            target_date.strftime("%Y-%m"),
                            target_date.strftime("%Y-%m-%d.txt"))
        lo = min(target_date, date.max - DAY)
        hi = lo + DAY
        return path, lo, hi

    def __iter__(self):
        """Return an iterator which yields all items in the data store
        sequntially. Equivalent to: for item in dataset[:]: yield item"""
        for item in self[:]: yield item

    def update(self, E):
        """D.update(E) -> None. Equivelent to: for k in E: D[ k['idx'] ] = k"

        Update D from list-like iterable E containing dicts.
        Pre-existing items being overwritten.
        Dicts are assumed to contain all appropriate keys and values."""
        for k in E:
            self[ k['idx'] ] = k

    def clear(self):
        """Clears all data from the data store permanently"""
        for root, dirs, files in os.walk(self._root_dir, topdown=False):
            for file in files:
                os.unlink(os.path.join(root, file))
            os.rmdir(root)
        # Get the root dir back and re-initialise to start again
        root_dir = os.path.abspath(
            os.path.join(self._root_dir, os.pardir))
        self.__init__(root_dir)

class RawStore(CoreStore):
    """Stores raw weather station data."""
    dir_name = 'raw'
    key_list = [
        'idx', 'delay', 'hum_in', 'temp_in', 'hum_out', 'temp_out',
        'abs_pressure', 'wind_ave', 'wind_gust', 'wind_dir', 'rain',
        'status', 'illuminance', 'uv',
        ]
    solar_items = 2
    conv = {
        'idx'          : WSDateTime.from_csv,
        'delay'        : int,
        'hum_in'       : int,
        'temp_in'      : float,
        'hum_out'      : int,
        'temp_out'     : float,
        'abs_pressure' : float,
        'wind_ave'     : float,
        'wind_gust'    : float,
        'wind_dir'     : int,
        'rain'         : float,
        'status'       : WSStatus.from_csv,
        'illuminance'  : float,
        'uv'           : int,
        }


class CalibStore(CoreStore):
    """Stores 'calibrated' weather station data."""
    dir_name = 'calib'
    key_list = [
        'idx', 'delay', 'hum_in', 'temp_in', 'hum_out', 'temp_out',
        'abs_pressure', 'rel_pressure', 'wind_ave', 'wind_gust', 'wind_dir',
        'rain', 'status', 'illuminance', 'uv',
        ]
    solar_items = 2
    conv = {
        'idx'          : WSDateTime.from_csv,
        'delay'        : int,
        'hum_in'       : int,
        'temp_in'      : float,
        'hum_out'      : int,
        'temp_out'     : float,
        'abs_pressure' : float,
        'rel_pressure' : float,
        'wind_ave'     : float,
        'wind_gust'    : float,
        'wind_dir'     : float,
        'rain'         : float,
        'status'       : WSStatus.from_csv,
        'illuminance'  : float,
        'uv'           : int,
        }


class HourlyStore(CoreStore):
    """Stores hourly summary weather station data."""
    dir_name = 'hourly'
    key_list = [
        'idx', 'hum_in', 'temp_in', 'hum_out', 'temp_out',
        'abs_pressure', 'rel_pressure', 'pressure_trend',
        'wind_ave', 'wind_gust', 'wind_dir', 'rain', 'illuminance', 'uv',
        ]
    solar_items = 2
    conv = {
        'idx'               : WSDateTime.from_csv,
        'hum_in'            : int,
        'temp_in'           : float,
        'hum_out'           : int,
        'temp_out'          : float,
        'abs_pressure'      : float,
        'rel_pressure'      : float,
        'pressure_trend'    : float,
        'wind_ave'          : float,
        'wind_gust'         : float,
        'wind_dir'          : float,
        'rain'              : float,
        'illuminance'       : float,
        'uv'                : int,
        }


class DailyStore(CoreStore):
    """Stores daily summary weather station data."""
    dir_name = 'daily'
    key_list = [
        'idx', 'start',
        'hum_out_ave',
        'hum_out_min', 'hum_out_min_t', 'hum_out_max', 'hum_out_max_t',
        'temp_out_ave',
        'temp_out_min', 'temp_out_min_t', 'temp_out_max', 'temp_out_max_t',
        'hum_in_ave',
        'hum_in_min', 'hum_in_min_t', 'hum_in_max', 'hum_in_max_t',
        'temp_in_ave',
        'temp_in_min', 'temp_in_min_t', 'temp_in_max', 'temp_in_max_t',
        'abs_pressure_ave',
        'abs_pressure_min', 'abs_pressure_min_t',
        'abs_pressure_max', 'abs_pressure_max_t',
        'rel_pressure_ave',
        'rel_pressure_min', 'rel_pressure_min_t',
        'rel_pressure_max', 'rel_pressure_max_t',
        'wind_ave', 'wind_gust', 'wind_gust_t', 'wind_dir',
        'rain',
        'illuminance_ave', 'illuminance_max', 'illuminance_max_t',
        'uv_ave', 'uv_max', 'uv_max_t',
        ]
    solar_items = 6
    conv = {
        'idx'                : WSDateTime.from_csv,
        'start'              : WSDateTime.from_csv,
        'hum_out_ave'        : float,
        'hum_out_min'        : int,
        'hum_out_min_t'      : WSDateTime.from_csv,
        'hum_out_max'        : int,
        'hum_out_max_t'      : WSDateTime.from_csv,
        'temp_out_ave'       : float,
        'temp_out_min'       : float,
        'temp_out_min_t'     : WSDateTime.from_csv,
        'temp_out_max'       : float,
        'temp_out_max_t'     : WSDateTime.from_csv,
        'hum_in_ave'         : float,
        'hum_in_min'         : int,
        'hum_in_min_t'       : WSDateTime.from_csv,
        'hum_in_max'         : int,
        'hum_in_max_t'       : WSDateTime.from_csv,
        'temp_in_ave'        : float,
        'temp_in_min'        : float,
        'temp_in_min_t'      : WSDateTime.from_csv,
        'temp_in_max'        : float,
        'temp_in_max_t'      : WSDateTime.from_csv,
        'abs_pressure_ave'   : float,
        'abs_pressure_min'   : float,
        'abs_pressure_min_t' : WSDateTime.from_csv,
        'abs_pressure_max'   : float,
        'abs_pressure_max_t' : WSDateTime.from_csv,
        'rel_pressure_ave'   : float,
        'rel_pressure_min'   : float,
        'rel_pressure_min_t' : WSDateTime.from_csv,
        'rel_pressure_max'   : float,
        'rel_pressure_max_t' : WSDateTime.from_csv,
        'wind_ave'           : float,
        'wind_gust'          : float,
        'wind_gust_t'        : WSDateTime.from_csv,
        'wind_dir'           : float,
        'rain'               : float,
        'illuminance_ave'    : float,
        'illuminance_max'    : float,
        'illuminance_max_t'  : WSDateTime.from_csv,
        'uv_ave'             : float,
        'uv_max'             : int,
        'uv_max_t'           : WSDateTime.from_csv,
        }

    def _get_cache_path(self, target_date):
        # one file per month
        path = os.path.join(self._root_dir,
                            target_date.strftime("%Y"),
                            target_date.strftime("%Y-%m-01.txt"))
        lo = target_date.replace(day=1)
        if lo.month < 12:
            hi = lo.replace(month=lo.month+1)
        elif lo.year < MAXYEAR:
            hi = lo.replace(year=lo.year+1, month=1)
        else:
            hi = lo
            lo = hi.replace(month=hi.month-1)
        return path, lo, hi


class MonthlyStore(CoreStore):
    """Stores monthly summary weather station data."""
    dir_name = 'monthly'
    key_list = [
        'idx', 'start',
        'hum_out_ave',
        'hum_out_min', 'hum_out_min_t', 'hum_out_max', 'hum_out_max_t',
        'temp_out_ave',
        'temp_out_min_lo', 'temp_out_min_lo_t',
        'temp_out_min_hi', 'temp_out_min_hi_t', 'temp_out_min_ave',
        'temp_out_max_lo', 'temp_out_max_lo_t',
        'temp_out_max_hi', 'temp_out_max_hi_t', 'temp_out_max_ave',
        'hum_in_ave',
        'hum_in_min', 'hum_in_min_t', 'hum_in_max', 'hum_in_max_t',
        'temp_in_ave',
        'temp_in_min_lo', 'temp_in_min_lo_t',
        'temp_in_min_hi', 'temp_in_min_hi_t', 'temp_in_min_ave',
        'temp_in_max_lo', 'temp_in_max_lo_t',
        'temp_in_max_hi', 'temp_in_max_hi_t', 'temp_in_max_ave',
        'abs_pressure_ave',
        'abs_pressure_min', 'abs_pressure_min_t',
        'abs_pressure_max', 'abs_pressure_max_t',
        'rel_pressure_ave',
        'rel_pressure_min', 'rel_pressure_min_t',
        'rel_pressure_max', 'rel_pressure_max_t',
        'wind_ave', 'wind_gust', 'wind_gust_t', 'wind_dir',
        'rain', 'rain_days',
        'illuminance_ave',
        'illuminance_max_lo', 'illuminance_max_lo_t',
        'illuminance_max_hi', 'illuminance_max_hi_t', 'illuminance_max_ave',
        'uv_ave',
        'uv_max_lo', 'uv_max_lo_t', 'uv_max_hi', 'uv_max_hi_t', 'uv_max_ave',
        ]
    solar_items = 12
    conv = {
        'idx'                  : WSDateTime.from_csv,
        'start'                : WSDateTime.from_csv,
        'hum_out_ave'          : float,
        'hum_out_min'          : int,
        'hum_out_min_t'        : WSDateTime.from_csv,
        'hum_out_max'          : int,
        'hum_out_max_t'        : WSDateTime.from_csv,
        'temp_out_ave'         : float,
        'temp_out_min_lo'      : float,
        'temp_out_min_lo_t'    : WSDateTime.from_csv,
        'temp_out_min_hi'      : float,
        'temp_out_min_hi_t'    : WSDateTime.from_csv,
        'temp_out_min_ave'     : float,
        'temp_out_max_lo'      : float,
        'temp_out_max_lo_t'    : WSDateTime.from_csv,
        'temp_out_max_hi'      : float,
        'temp_out_max_hi_t'    : WSDateTime.from_csv,
        'temp_out_max_ave'     : float,
        'hum_in_ave'           : float,
        'hum_in_min'           : int,
        'hum_in_min_t'         : WSDateTime.from_csv,
        'hum_in_max'           : int,
        'hum_in_max_t'         : WSDateTime.from_csv,
        'temp_in_ave'          : float,
        'temp_in_min_lo'       : float,
        'temp_in_min_lo_t'     : WSDateTime.from_csv,
        'temp_in_min_hi'       : float,
        'temp_in_min_hi_t'     : WSDateTime.from_csv,
        'temp_in_min_ave'      : float,
        'temp_in_max_lo'       : float,
        'temp_in_max_lo_t'     : WSDateTime.from_csv,
        'temp_in_max_hi'       : float,
        'temp_in_max_hi_t'     : WSDateTime.from_csv,
        'temp_in_max_ave'      : float,
        'abs_pressure_ave'     : float,
        'abs_pressure_min'     : float,
        'abs_pressure_min_t'   : WSDateTime.from_csv,
        'abs_pressure_max'     : float,
        'abs_pressure_max_t'   : WSDateTime.from_csv,
        'rel_pressure_ave'     : float,
        'rel_pressure_min'     : float,
        'rel_pressure_min_t'   : WSDateTime.from_csv,
        'rel_pressure_max'     : float,
        'rel_pressure_max_t'   : WSDateTime.from_csv,
        'wind_ave'             : float,
        'wind_gust'            : float,
        'wind_gust_t'          : WSDateTime.from_csv,
        'wind_dir'             : float,
        'rain'                 : float,
        'rain_days'            : int,
        'illuminance_ave'      : float,
        'illuminance_max_lo'   : float,
        'illuminance_max_lo_t' : WSDateTime.from_csv,
        'illuminance_max_hi'   : float,
        'illuminance_max_hi_t' : WSDateTime.from_csv,
        'illuminance_max_ave'  : float,
        'uv_ave'               : float,
        'uv_max_lo'            : int,
        'uv_max_lo_t'          : WSDateTime.from_csv,
        'uv_max_hi'            : int,
        'uv_max_hi_t'          : WSDateTime.from_csv,
        'uv_max_ave'           : float,
        }

    def _get_cache_path(self, target_date):
        # one file per year
        path = os.path.join(self._root_dir,
                            target_date.strftime("%Y-01-01.txt"))
        lo = target_date.replace(month=1, day=1)
        if lo.year < MAXYEAR:
            hi = lo.replace(year=lo.year+1)
        else:
            hi = lo
            lo = hi.replace(year=hi.year-1)
        return path, lo, hi

